// Copyright (c) 2009-2010 Satoshi Nakamoto
// Copyright (c) 2009-2016 The Bitcoin Core developers
// Distributed under the MIT software license, see the accompanying
// file COPYING or http://www.opensource.org/licenses/mit-license.php.

#include "pow.h"

#include "arith_uint256.h"
#include "chain.h"
#include "primitives/block.h"
#include "uint256.h"
//TOG(Triger of Block Timeout Generation)
#include "utiltime.h"

//BTF Network Difficulty ctl:BtfDiffEnable setup for 212700 forking and init a new Difficulty for all new miners..
static bool BtfDiffEnable(const CBlockIndex* pindexPrev){
 const int nHeight = pindexPrev->nHeight + 1;
    if (nHeight>=212701){
        return true;
    }else if(nHeight<=212700){
        return false;
    }else{
        return true;
    }
}


//Bitcoin Finance Network: TOG(Triger of Block Timeout Generation)
    /*
    This checker is used for invalid block attacks, unless the majority of BTF nodes are located in a system time less than NTP time, which does not exist. Therefore, attackers cannot commit fraud by creating an advanced block. If the attacker creates the attack block in advance and broadcasts it at the arrival time, this possibility exists because it complies with the TOG timeout rule. Therefore an additional 3000ms gap is added to avoid this situation.
    */
static bool trigerOfBlockGeneration(const CBlockIndex* pindexLast, const CBlockHeader *pblock ){
    const  int64_t sysStp =  GetSystemTimeInSeconds();
    //Gap 3000 ms for global TCP delay
    if (sysStp + 3  < pblock->GetBlockTime()){
         std::cout<<"BitcoinFinance Network ? TOG(Triger of Block Timeout Generation) -> Warning: Invalid block time or system time not synchronized to NTP service. "<<"\n";
         return false;
    }else if( (pblock->GetBlockTime() -  pindexLast->GetBlockTime() >= 1980)  &&  (sysStp - pindexLast->GetBlockTime() >=  1980)){
         std::cout<<"BitcoinFinance Network <-  TOG been Triggered  <- TOG(Triger of Block Timeout Generation)  "<<"\n";
        return true;
    }else{
        return false;
    }
}

 


/*
Check Point of TOG(Triger of Block Timeout Generation) 
*/
static bool trigerBlockCheckPoint(const CBlockIndex* pindexPrev){
 const int nHeight = pindexPrev->nHeight + 1;
    if (nHeight>=261210){
        return true;
    }else if(nHeight<=261209){
        return false;
    }else{
        return true;
    }
}


static bool upboost(const CBlockIndex* pindexPrev){
 const int nHeight = pindexPrev->nHeight + 1;
    if (nHeight>=267960){
        return true;
    }else if(nHeight<=267959){
        return false;
    }else{
        return true;
    }
}

//For TOG Type2: TOG-FORCE Calculate Difficulty base on Last 12 Blocks
static bool ckptTendency(const CBlockIndex* pindexPrev){
 const int nHeight = pindexPrev->nHeight + 1;
    if (nHeight>=269269){
        return true;
    }else if(nHeight<=269268){
        return false;
    }else{
        return true;
    }
}


unsigned int GetNextWorkRequired(const CBlockIndex* pindexLast, const CBlockHeader *pblock, const Consensus::Params& params)
{
    assert(pindexLast != nullptr);
    unsigned int nProofOfWorkLimit = UintToArith256(params.powLimit).GetCompact();
    if (!BtfDiffEnable(pindexLast)){
        std::cout<<"BTF Network Difficulty ctl:BtfDiffEnable disabled for BTC chain forking."<<"\n";
        return pindexLast->nBits;
    }
    //Bitcoin Finance Network: TOG(Triger of Block Timeout Generation)
     if (trigerBlockCheckPoint(pindexLast)){   //CheckPoint of TOG >=261210
            if (trigerOfBlockGeneration(pindexLast,pblock)){
                        if (ckptTendency(pindexLast)==false){
                                //Type1: TOG-NORMAL
                                //Calculate  TOG(Triger of Block Timeout Generation) NextWork Required immediately,Only works while Block.Num >= 261210 and
                                //reach Too high difficulty for mining and Hashrate.
                                        //BTF network's params.DifficultyAdjustmentInterval()=2016 blocks
                                        int nHeightFirst = pindexLast->nHeight - (params.DifficultyAdjustmentInterval()-1);
                                        //int nHeightFirst = pindexLast->nHeight - (2-1);  //BTF Network debug remarked 2block -> adjustment
                                        assert(nHeightFirst >= 0);
                                        const CBlockIndex* pindexFirst = pindexLast->GetAncestor(nHeightFirst);
                                        assert(pindexFirst);
                                        std::cout<<"BitcoinFinance Network <- TOG-NORMAL  <- TOG(Triger of Block Timeout Generation)  #ON"<<"\n";
                                        return CalculateNextWorkRequired(pindexLast, pindexFirst->GetBlockTime(), params);
                        }else{
                                //Type2: TOG-FORCE Calculate Difficulty base on Last 12 Blocks
                                        //Type2 = 2016/168 = 12  Using 1/168 of DifficultyAdjustmentInterval()=2016 blocks Calculate new Difficulty
                                        int nHeightFirst = pindexLast->nHeight -  (((params.DifficultyAdjustmentInterval())/168)-1);
                                        assert(nHeightFirst >= 0);
                                        const CBlockIndex* pindexFirst = pindexLast->GetAncestor(nHeightFirst);
                                        assert(pindexFirst);
                                        int64_t reMainTimeSpan = pindexLast->GetBlockTime() - pindexFirst->GetBlockTime();  //last block timestamp - 12 blocks before timestamp
                                        // 14 * 24 * 60 * 60 / 168 = 7200S Checking for TOG type
                                        std::cout<<"Type2: TOG-FORCE Calculate Difficulty ->  [reMainTimeSpan]:"<<reMainTimeSpan<<"\n";
                                        if (reMainTimeSpan>=7200){
                                            //blocks slow  BitcoinFinance Network Difficulty Adjustment in TOG in Type1
                                            std::cout<<"BitcoinFinance Network <-TOG-Type1: TOG-FORCE Calculate Difficulty base on Last 12 Blocks  <- TOG(Triger of Block Timeout Generation)  #ON"<<"\n";
                                            return CalculateNextWorkRequired(pindexLast, (pindexFirst->GetBlockTime()-(reMainTimeSpan*167)), params);
                                        }
                                        if ((reMainTimeSpan>=360)&&(reMainTimeSpan<7200)){
                                            //Pow decreased,Force BitcoinFinance Network Difficulty Adjustment in TOG Type2
                                            std::cout<<"BitcoinFinance Network <-TOG-Type2: TOG-FORCE Calculate Difficulty base on Last 12 Blocks  <- TOG(Triger of Block Timeout Generation)  #ON"<<"\n";
                                            // 7200/360 =20 , Pow decreased quick,So make First block's time big enough for  while  BitcoinFinance Network Difficulty Adjustment while TOG triged.
                                            // This setting may lead to premature difficulty reduction, but maintaining the stability of the entire network is more important than adjusting the difficulty too high
                                            return CalculateNextWorkRequired(pindexLast, (pindexFirst->GetBlockTime()-(reMainTimeSpan*167*20)), params);
                                        }
                                        if (reMainTimeSpan<360){
                                            //Block's time length<=30S ,may caused by Mining Pool issue or bitcoin finance network node upgrade  360/12=30s
                                            //There is still a possibility of error in this judgment, but it needs to be based on a very high Pow hash rate and quickly reduced in order to occur. In the case of a block triggering a TOG of less than 30 seconds, the difficulty of the block should not be adjusted temporarily to meet the needs of the mining pool.Definitely it is a very extreme situation, and we believe that maintaining the stability of the entire network and adapting to fluctuations in hash rates are important.
                                            std::cout<<"BitcoinFinance Network <- TOG-REMAIN DIFF  <- TOG(Triger of Block Timeout Generation)  #CANCELED AS RETURN(pindexLast->nBits)"<<"\n";
                                           return pindexLast->nBits;
                                        }
                        }
            }
     }
     
     
    // Only change once per difficulty adjustment interval
    //if ((pindexLast->nHeight+1) % 2 != 0) 
    if ((pindexLast->nHeight+1) % params.DifficultyAdjustmentInterval() != 0)
    {   
        std::cout<<"Bitcoin Finance (BTF) Network: New Block mined found - - - > Notify."<<"\n";
        if (params.fPowAllowMinDifficultyBlocks)
        {
            // Special difficulty rule for testnet:
            std::cout<<"BTF Network: Special difficulty rule for testnet - > Notify."<<"\n";
            // If the new block's timestamp is more than 2* 10 minutes
            // then allow mining of a min-difficulty block.
            if (pblock->GetBlockTime() > pindexLast->GetBlockTime() + params.nPowTargetSpacing*2)
                return nProofOfWorkLimit;
            else
            {
                // Return the last non-special-min-difficulty-rules-block  
                const CBlockIndex* pindex = pindexLast;
                while (pindex->pprev && pindex->nHeight % params.DifficultyAdjustmentInterval() != 0 && pindex->nBits == nProofOfWorkLimit)
                    pindex = pindex->pprev;
                    std::cout<<"BTF Network:Return the last non-special-min-difficulty-rules-block ."<<"\n";
                return pindex->nBits;
            }
        }

        return pindexLast->nBits;
    }

    // Go back by what we want to be 14 days worth of blocks
    //BTF network's params.DifficultyAdjustmentInterval()=2016 blocks
    int nHeightFirst = pindexLast->nHeight - (params.DifficultyAdjustmentInterval()-1);
    //int nHeightFirst = pindexLast->nHeight - (2-1);  //BTF Network debug remarked 2block -> adjustment
    assert(nHeightFirst >= 0);
    const CBlockIndex* pindexFirst = pindexLast->GetAncestor(nHeightFirst);
    assert(pindexFirst);
    std::cout<<"BTC ->  #[CalculateTogRequired] -> EXTEND ->Bitcoin Finance(BTF) network's params.DifficultyAdjustmentInterval Retargeting -> Retarget."<<"\n";
    return CalculateNextWorkRequired(pindexLast, pindexFirst->GetBlockTime(), params);
}


unsigned int CalculateNextWorkRequired(const CBlockIndex* pindexLast, int64_t nFirstBlockTime, const Consensus::Params& params)
{
            if (params.fPowNoRetargeting)
                return pindexLast->nBits;
           
            int64_t nActualTimespan = pindexLast->GetBlockTime() - nFirstBlockTime;
            std::cout<<"pindexLast->GetBlockTime():"<<(pindexLast->GetBlockTime())<<"\n";
            std::cout<<"nFirstBlockTime:"<<nFirstBlockTime<<"\n";
            std::cout<<"(pindexLast-nFirstBlockTime) ->  [ActualTimespan]:"<<nActualTimespan<<"\n";
            //Too Fast -> bitcoin default blocks <267960 
            if ((nActualTimespan < params.nPowTargetTimespan/4)&&(upboost(pindexLast)==false)){ //ActualTimespan < 3.5 days
                std::cout<<"BTC: BLOCK TOO FAST ->   #[ActualTimespan]:"<<nActualTimespan<<"\n";
                nActualTimespan = params.nPowTargetTimespan/4;
            std::cout<<"BTC ->  #[CalculateTogRequired] -> Bitcoin Finance EXTEND -> ActualTimespan/4 RTG BTC Default -> (Bitcoin Core Default:ActualTimespan < 3.5 days)"<<"\n";
            }
            //blocks  >=267960  increase difficulty only once
            if ((nActualTimespan < params.nPowTargetTimespan/4)&&(upboost(pindexLast)==true)&&((pindexLast->nHeight+1) % params.DifficultyAdjustmentInterval() == 0)){ //ActualTimespan < 3.5 days
                std::cout<<"BTC: BLOCK TOO FAST ->   #[ActualTimespan]:"<<nActualTimespan<<"\n";
                nActualTimespan = params.nPowTargetTimespan/4;
            std::cout<<"BTC ->  #[CalculateTogRequired] -> Bitcoin Finance EXTEND -> ActualTimespan/4 RTG BTC Default -> (Bitcoin Core Default:ActualTimespan < 3.5 days)"<<"\n";
            }
            
            //blocks  >=267960 & not triged  params.DifficultyAdjustmentInterval 2016 
            if ((nActualTimespan < params.nPowTargetTimespan/4)&&(upboost(pindexLast)==true)&&((pindexLast->nHeight+1) % params.DifficultyAdjustmentInterval() != 0)){ //ActualTimespan < 3.5 days
                std::cout<<"BTC: BLOCK TOO FAST first then Slower -> nBit not change  #[ActualTimespan]:"<<nActualTimespan<<"\n";
                return pindexLast->nBits;
            }
            
            //Too slow -> bitcoin default
            if ((nActualTimespan >= params.nPowTargetTimespan*4)&&(nActualTimespan < params.nPowTargetTimespan*5)){  //ActualTimespan  56 ~ 70 days
                std::cout<<"BTC: BLOCK TOO SLOW ->  #[ActualTimespan]:"<<nActualTimespan<<"\n";
                nActualTimespan = params.nPowTargetTimespan*4;
                std::cout<<"BTC ->  #[CalculateTogRequired] -> Bitcoin Finance EXTEND -> ActualTimespan*4 BTC Default -> (Bitcoin Core Default:ActualTimespan > 56 days)"<<"\n";
            }
             // Too long block pending -> diff / 4k -> New diff
             
            if( (nActualTimespan >= params.nPowTargetTimespan*5)&&(nActualTimespan < params.nPowTargetTimespan*6)){  //ActualTimespan > 70 ~ 84days
                std::cout<<"BTC ->   #[ActualTimespan]:"<<nActualTimespan<<"\n";
                nActualTimespan = params.nPowTargetTimespan*100000;
                std::cout<<"BTC ->  #[CalculateTogRequired] -> Bitcoin Finance TOG-> ActualTimespan*100K BTC_RTG -> TOG-ON(Long Block Pending:ActualTimespan  70~84 days)"<<"\n";
            }

            // Too long block pending -> diff / 400k -> New diff
            if (nActualTimespan >= params.nPowTargetTimespan*6){  //ActualTimespan > 84 days
                std::cout<<"BTC ->  #[ActualTimespan]:"<<nActualTimespan<<"\n";
                nActualTimespan = params.nPowTargetTimespan*4000000;
                std::cout<<"BTC ->  #[CalculateTogRequired] -> Bitcoin Finance TOG-> ActualTimespan*400K BTC_RTG -> TOG-ON(Long Block Pending:ActualTimespan > 84 days)"<<"\n";
            }

            // Retarget
            const arith_uint256 bnPowLimit = UintToArith256(params.powLimit);
            const arith_uint256 btfFlexPow = UintToArith256(uint256S("0000000000000000000000000000000000000000000000000000000000000003"));
            arith_uint256 bnNew;
            bnNew.SetCompact(pindexLast->nBits);
            if (bnNew<=btfFlexPow){
                //avoid zerx difficulty in Btf network forks
                const arith_uint256 btfInitedPow = UintToArith256(uint256S("0000000000000000000000000000000000000000000000000000000000000003"));
                bnNew *= nActualTimespan*btfInitedPow;
                bnNew /= params.nPowTargetTimespan;
                bnNew +=btfInitedPow;
                std::cout<<"BTC ->  #[CalculateTogRequired] -> Void zero difficulty"<<"\n";
            }else{
                //Start a new difficulty regarting
                std::cout<<"BTC ->  T1#[nActualTimespan]:"<<nActualTimespan<<"\n";
                bnNew *= nActualTimespan;               //
                bnNew /= params.nPowTargetTimespan;   //1209600
                std::cout <<"BTC ->   #[params.nPowTargetTimespan]:"<< params.nPowTargetTimespan<<"\n";
                std::cout <<"BTC ->   #[CalculateTogRequired] -> new difficulty regarting setup"<<"\n";
            }
            if (bnNew > bnPowLimit){
                bnNew = bnPowLimit;
                std::cout<<"BTC ->  #[CalculateTogRequired] -> new difficulty reached Btf.bnPowLimit -> setup (bnNew = bnPowLimit) -> GetCompacted"<<"\n";
            }
            return bnNew.GetCompact();
}


/*

unsigned int CalculateNextWorkRequired(const CBlockIndex* pindexLast, int64_t nFirstBlockTime, const Consensus::Params& params)
{
            if (params.fPowNoRetargeting)
                return pindexLast->nBits;
            // Limit adjustment step
            int64_t nActualTimespan = pindexLast->GetBlockTime() - nFirstBlockTime;
            //setting nActualTimespan-> 2016 = 168 * 12 = 2 * 1008
            //nActualTimespan*=1008; //BTF Network debug remarked 2*1008=2016 -> adjustment debug
            if (nActualTimespan < params.nPowTargetTimespan/4){
                nActualTimespan = params.nPowTargetTimespan/4;
            }
            if (nActualTimespan > params.nPowTargetTimespan*4){
                nActualTimespan = params.nPowTargetTimespan*4;
            }
            // Retarget
            const arith_uint256 bnPowLimit = UintToArith256(params.powLimit);
            const arith_uint256 btfFlexPow = UintToArith256(uint256S("0000000000000000000000000000000000000000000000000000000000000003"));
            arith_uint256 bnNew;
            bnNew.SetCompact(pindexLast->nBits);
            if (bnNew<=btfFlexPow){
                //avoid zerx difficulty in Btf network forks
            const arith_uint256 btfInitedPow = UintToArith256(uint256S("0000000000000000000000000000000000000000000000000000000000000003"));
                bnNew *= nActualTimespan*btfInitedPow;
                bnNew /= params.nPowTargetTimespan;
                bnNew +=btfInitedPow;
                std::cout<<"Bitcoin fork  -> EXTEND -> Bitcoin Finance(BTF) network difficulty btfInitedPow inited -> adjustment [+3]."<<"\n";
            }else{
                //Start a new difficulty regarting
                bnNew *= nActualTimespan;               //
                bnNew /= params.nPowTargetTimespan;   //1209600
                std::cout<<"Bitcoin fork -> EXTEND -> Bitcoin Finance(BTF) network -> new difficulty regarting setup."<<"\n";
            }
            if (bnNew > bnPowLimit){
                bnNew = bnPowLimit;
                std::cout<<"Bitcoin fork -> EXTEND -> Bitcoin Finance(BTF) network -> new difficulty reached Btf.bnPowLimit -> setup (bnNew = bnPowLimit)."<<"\n";
            }
            return bnNew.GetCompact();
}

*/

bool CheckProofOfWork(uint256 hash, unsigned int nBits, const Consensus::Params& params)
{
    bool fNegative;
    bool fOverflow;
    arith_uint256 bnTarget;

    bnTarget.SetCompact(nBits, &fNegative, &fOverflow);

    // Check range
    if (fNegative || bnTarget == 0 || fOverflow || bnTarget > UintToArith256(params.powLimit))
        return false;

    // Check proof of work matches claimed amount
    if (UintToArith256(hash) > bnTarget)
        return false;

    return true;
}
